/*
 * This file was written by Milan Straka as a part of iqc library.
 * Copyright (C) 2008 Milan Straka.
 * 
 * This program is free software: you can redistribute it and/or 
 * modify it under the terms of the GNU Lesser General Public 
 * License as published by the Free Software Foundation, either 
 * version 3 of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General 
 * Public License along with this program.  If not, see 
 * <http://www.gnu.org/licenses/>.
 */

#ifndef LIBIQC_BY_MILAN_STRAKA_2008
#define LIBIQC_BY_MILAN_STRAKA_2008

// Basic definitions
#include <stdint.h>
#include <gmp.h>

typedef struct iqc_struct {
  mpz_t a,b;
} iqc_t[1];

void iqc_set_disc(mpz_t d);
void iqc_get_disc(mpz_t d);

#define iqc_init(x)   do { mpz_init(x->a);      mpz_init(x->b);     } while (0)
#define iqc_clear(x)  do { mpz_clear(x->a);     mpz_clear(x->b);    } while (0)
#define iqc_set(x,A)  do { mpz_set(x->a,A->a);  mpz_set(x->b,A->b); } while (0)
#define iqc_equal(x,y)   (!mpz_cmp(x->a,y->a)&&!mpz_cmp(x->b,y->b))

//Disc generation
void iqc_gen_disc1(mpz_t d, unsigned long bits);
void iqc_gen_disc2(mpz_t d, unsigned long bits);
void iqc_set_gen_disc1(unsigned long bits);
void iqc_set_gen_disc2(unsigned long bits);

//Operations
void iqc_identity_basic(iqc_t x);
void iqc_invert_basic(iqc_t x, iqc_t a);
void iqc_reduce_basic(iqc_t x, iqc_t a);
void iqc_mul_basic(iqc_t x, iqc_t a, iqc_t b);
void iqc_sqr_basic(iqc_t x, iqc_t a);
void iqc_mul_nucomp(iqc_t x, iqc_t a, iqc_t b);
void iqc_sqr_nucomp(iqc_t x, iqc_t a);
void iqc_exp_basic(iqc_t x, iqc_t a, mpz_t e);
void iqc_exp_signed(iqc_t x, iqc_t a, mpz_t e);
void iqc_random(iqc_t x);
#define iqc_identity(x) iqc_identity_basic(x)
#define iqc_invert(x,a) iqc_invert_basic(x,a)
#define iqc_reduce(x,a) iqc_reduce_basic(x,a)
#define iqc_mul(x,a,b)  iqc_mul_nucomp(x,a,b)
#define iqc_sqr(x,a)    iqc_sqr_nucomp(x,a)
#define iqc_exp(x,a,e)  iqc_exp_signed(x,a,e)

//Protocols - hash interface
#define HASH_DIGEST_LEN 20
typedef struct SHA1Context
{
    uint32_t Intermediate_Hash[HASH_DIGEST_LEN/4]; /*Message Digest*/
    uint32_t Length_Low;            /* Message length in bits      */
    uint32_t Length_High;           /* Message length in bits      */
                               /* Index into message block array   */
    int_least16_t Message_Block_Index;
    uint8_t Message_Block[64];      /* 512-bit message blocks      */
    int Computed;               /* Is the digest computed?         */
    int Corrupted;             /* Is the message digest corrupted? */
} SHA1Context;

typedef struct hash_state_struct {
  SHA1Context sha;
} hash_state_t[1];

void hash_init(hash_state_t s);
void hash_data(hash_state_t s, void* ptr, unsigned long len);
void hash_iqc(hash_state_t s, iqc_t a);
void hash_finish(hash_state_t s, unsigned char digest[HASH_DIGEST_LEN]);

#define SERIALIZABLE_INIT(name)  void name##_init(name##_t what)
#define SERIALIZABLE_CLER(name)  void name##_clear(name##_t what)
#define SERIALIZABLE_LOAD(name)  void name##_load(name##_t what, const char* fname)
#define SERIALIZABLE_SAVE(name)  void name##_save(name##_t what, const char* fname)
#define SERIALIZABLE(name)       SERIALIZABLE_INIT(name); SERIALIZABLE_CLER(name); \
                                 SERIALIZABLE_LOAD(name); SERIALIZABLE_SAVE(name)

//Protocols - elgamal
#define ELGAMAL_ENC_LEN HASH_DIGEST_LEN

typedef struct elgamal_pubk_struct  { mpz_t D; iqc_t c,a;                        } elgamal_pubk_t[1];  
typedef struct elgamal_privk_struct { mpz_t a;                                   } elgamal_privk_t[1]; 
typedef struct elgamal_enc_struct   { iqc_t g; unsigned char C[ELGAMAL_ENC_LEN]; } elgamal_enc_t[1];   
SERIALIZABLE(elgamal_pubk);  SERIALIZABLE(elgamal_privk);  SERIALIZABLE(elgamal_enc);

void elgamal_gen(int bits, elgamal_pubk_t pub, elgamal_privk_t priv);
void elgamal_enc(elgamal_pubk_t pub, unsigned char M[ELGAMAL_ENC_LEN], elgamal_enc_t enc);
void elgamal_dec(elgamal_pubk_t pub, elgamal_privk_t priv,elgamal_enc_t enc,unsigned char M[ELGAMAL_ENC_LEN]);

//Protocols - gq
typedef struct gq_pubk_struct  { mpz_t D,q; iqc_t a;                        } gq_pubk_t[1];
typedef struct gq_privk_struct { iqc_t a;                                   } gq_privk_t[1];
typedef struct gq_sig_struct   { iqc_t w; unsigned char s[HASH_DIGEST_LEN]; } gq_sig_t[1];
SERIALIZABLE(gq_pubk);  SERIALIZABLE(gq_privk);  SERIALIZABLE(gq_sig);

void gq_gen(int bits, gq_pubk_t pub, gq_privk_t priv);
void gq_sig(gq_pubk_t pub, gq_privk_t priv, hash_state_t doc, gq_sig_t sig);
int  gq_ver(gq_pubk_t pub, hash_state_t doc, gq_sig_t sig);

//Protocols - gps
typedef struct gps_pubk_struct  { mpz_t D; iqc_t c,a; } gps_pubk_t[1];
typedef struct gps_privk_struct { mpz_t a;            } gps_privk_t[1];
typedef struct gps_sig_struct   { mpz_t s; iqc_t r;   } gps_sig_t[1];
SERIALIZABLE(gps_pubk);  SERIALIZABLE(gps_privk);  SERIALIZABLE(gps_sig);

void gps_gen(int bits, gps_pubk_t pub, gps_privk_t priv);
void gps_sig(gps_pubk_t pub, gps_privk_t priv, hash_state_t doc, gps_sig_t sig);
int  gps_ver(gps_pubk_t pub, hash_state_t doc, gps_sig_t sig);

#endif
