/*
 * This file was written by Milan Straka as a part of iqc library.
 * Copyright (C) 2008 Milan Straka.
 * 
 * This program is free software: you can redistribute it and/or 
 * modify it under the terms of the GNU Lesser General Public 
 * License as published by the Free Software Foundation, either 
 * version 3 of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General 
 * Public License along with this program.  If not, see 
 * <http://www.gnu.org/licenses/>.
 */

#include "oper_basic.h"

//Constructors
static mpz_t red_c, red_q, red_r, red_t, mul_d, mul_u, mul_v, mul_t, mul_b;
IQC_CONSTRUCTOR(oper_basic) {
  mpz_init(red_c); mpz_init(red_q); mpz_init(red_r); mpz_init(red_t);
  mpz_init(mul_d); mpz_init(mul_u); mpz_init(mul_v); mpz_init(mul_t); mpz_init(mul_b);
}

//Methods
void iqc_identity_basic(iqc_t x) {
  mpz_set_ui(x->a,1);
  mpz_set_ui(x->b,mpz_tstbit(iqc_disc, 0));
}

void iqc_invert_basic(iqc_t x,iqc_t a) {
  mpz_set(x->a,a->a);
  if (mpz_cmp(a->a,a->b)==0) mpz_set(x->b,a->b); else mpz_neg(x->b,a->b);
}

void iqc_reduce_lowlevel(iqc_t x, mpz_t a, mpz_t b, mpz_t c) {
#ifdef DEBUG
  mpz_mul(red_r, b, b);  mpz_mul(red_q, a, c);  mpz_mul_2exp(red_q, red_q, 2);
  mpz_sub(red_r, red_r, red_q);
  assert(mpz_cmp(iqc_disc, red_r)==0);
#endif

  mpz_set(x->a, a); mpz_set(x->b, b);
  if ((mpz_cmpabs(a,b)>0 || mpz_cmp(a,b)==0) && mpz_cmp(a,c) < 0) return;
  do {
    mpz_swap(x->a,c); mpz_neg(x->b,x->b);

    mpz_mul_2exp(red_t, x->a, 1);    
    mpz_fdiv_qr(red_q,red_r,x->b,red_t);
    if (mpz_cmp(red_r,x->a)>0) {mpz_sub(red_r,red_r,red_t); mpz_add_ui(red_q,red_q,1);}

    mpz_add(red_t,x->b,red_r); mpz_mul(red_t,red_t,red_q); mpz_tdiv_q_2exp(red_q,red_t,1);    
    mpz_sub(c,c,red_q);
    mpz_set(x->b,red_r);
  } while (mpz_cmp(x->a,c)>0);
  if (mpz_cmp(x->a,c)==0 && mpz_sgn(x->b)<0) mpz_neg(x->b,x->b);
}

void iqc_reduce_basic(iqc_t x, iqc_t a) {
#ifdef DEBUG
  assert(iqc_valid_form(a));
#endif

  mpz_neg(red_c, iqc_disc);            mpz_addmul(red_c, a->b, a->b);
  mpz_divexact(red_c, red_c, a->a);    mpz_tdiv_q_2exp(red_c, red_c, 2);
  iqc_reduce_lowlevel(x, a->a, a->b, red_c);
}


void iqc_mul_basic(iqc_t x, iqc_t a, iqc_t b) {
#ifdef DEBUG
  assert(iqc_valid_form(a));
#endif

  mpz_gcdext(mul_d,mul_u,mul_v,a->a,b->a);

  mpz_set(mul_b,b->b);
  mpz_add(mul_u,a->b,b->b);
  mpz_tdiv_q_2exp(mul_u,mul_u,1);
  mpz_sub(x->b,a->b,b->b);
  mpz_mul(x->b,x->b,b->a);
  mpz_mul(x->b,x->b,mul_v);
  mpz_mul(x->a,a->a,b->a);

  if (mpz_cmp_ui(mul_d,1)>0) {
    mpz_gcdext(mul_d,mul_u,mul_v,mul_d,mul_u);

    mpz_set(mul_t,iqc_disc);
    mpz_submul(mul_t,mul_b,mul_b);
    mpz_tdiv_q_2exp(mul_t,mul_t,1);
    mpz_mul(x->b,x->b,mul_u);
    mpz_addmul(x->b,mul_v,mul_t);
    mpz_divexact(x->b,x->b,mul_d);

    mpz_mul(mul_d,mul_d,mul_d);
    mpz_divexact(x->a,x->a,mul_d);
  }

  mpz_add(x->b,x->b,mul_b);
  mpz_mul_2exp(mul_t,x->a,1);
  mpz_mod(x->b,x->b,mul_t);

  iqc_reduce_basic(x,x);
}


void iqc_sqr_basic(iqc_t x, iqc_t a) {
#ifdef DEBUG
  assert(iqc_valid_form(a));
#endif

  mpz_gcdext(mul_d,mul_u,mul_v,a->a,a->b);
  mpz_divexact(x->a,a->a,mul_d);
  mpz_mul(x->a,x->a,x->a);

  mpz_set(mul_b, a->b);
  mpz_set(x->b,iqc_disc);
  mpz_submul(x->b,mul_b,mul_b);
  mpz_mul(x->b,x->b,mul_v);
  mpz_divexact(mul_t,x->b,mul_d);
  mpz_tdiv_q_2exp(x->b,mul_t,1);

  mpz_add(x->b,x->b,mul_b);
  mpz_mul_2exp(mul_t,x->a,1);
  mpz_mod(x->b,x->b,mul_t);

  iqc_reduce_basic(x,x);
}
